package ru.dbotthepony.kstarbound.client.render

import org.lwjgl.opengl.GL46
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kstarbound.client.gl.shader.GLShaderProgram
import ru.dbotthepony.kstarbound.client.gl.vertex.VertexBuilder
import ru.dbotthepony.kvector.arrays.Matrix4f

class Mesh(state: GLStateTracker) {
	constructor(state: GLStateTracker, builder: VertexBuilder) : this(state) {
		load(builder, GL46.GL_STATIC_DRAW)
	}

	val vao = state.newVAO()
	val vbo = state.newVBO()
	val ebo = state.newEBO()

	var indexCount = 0
		private set
	var indexType = 0
		private set

	fun load(builder: VertexBuilder, mode: Int = GL46.GL_DYNAMIC_DRAW) {
		vao.bind()
		vbo.bind()
		ebo.bind()

		builder.upload(vbo, ebo, mode)
		builder.attributes.apply(vao, true)

		indexCount = builder.indexCount
		indexType = builder.indexType

		vao.unbind()
		vbo.unbind()
		ebo.unbind()
	}

	fun render() {
		vao.bind()
		GL46.glDrawElements(GL46.GL_TRIANGLES, indexCount, indexType, 0L)
		checkForGLError()
		vao.unbind()
	}
}

data class ConfiguredMesh<T : GLShaderProgram>(val config: RenderConfig<T>, val mesh: Mesh = Mesh(config.state)) {
	fun render(transform: Matrix4f = config.state.matrixStack.last()) {
		config.setup(transform)
		mesh.render()
		config.uninstall()
	}
}
